package dev.peerat.parser.java.operation;

import java.util.List;
import java.util.function.Function;

import dev.peerat.parser.java.JavaElement;
import dev.peerat.parser.java.Operation;
import dev.peerat.parser.java.value.Value;
import dev.peerat.parser.java.visitor.JavaVisitor;
import dev.peerat.parser.visitor.VisitorBag;

public class ThrowOperation extends Operation{
	
	private Value value;
	
	public ThrowOperation(Value value){
		this.value = value;
	}
	
	public Value getValue(){
		return this.value;
	}
	
	@Override
	public void build(Builder builder) throws Exception {
		builder.append(" throw ");
		value.build(builder);
	}

	@Override
	public <E extends JavaElement> E find(Function<JavaElement, Boolean> finder) {
		return finder.apply(value) ? (E)value : null;
	}
	
	@Override
	public <E extends JavaElement> void findAll(Function<JavaElement, Boolean> finder, List<E> list){
		if(finder.apply(value)) list.add((E) value);
		value.findAll(finder, list);
	}

	@Override
	public VisitorBag visit(JavaVisitor<?> visitor) {
		if(visitor.canVisit(getClass())) return visitor.visit(this);
		VisitorBag bag = new VisitorBag();
		bag.merge(this.value.visit(visitor));
		return bag;
	}

}
