package dev.peerat.parser.java.visitor;

import dev.peerat.parser.java.JavaElement;
import dev.peerat.parser.java.visitor.operation.JavaOperationVisitor;
import dev.peerat.parser.java.visitor.value.JavaValueVisitor;
import dev.peerat.parser.visitor.Visitor;
import dev.peerat.parser.visitor.VisitorBag;

public abstract class JavaVisitor<T extends JavaElement> extends Visitor<JavaElement>{
	
	//TODO current system don't do litteraly ALL, juste DIRECT SUB ELEMENT
	
	public static <T extends JavaElement> JavaVisitor<T> base(Visitor<T> visitor){
		return new JavaVisitor<T>(null){
			
			@Override
			public boolean canVisit(Class<?> type){
//				System.out.println("base can visit ? "+type);
				return visitor.canVisit(type);
			}
			
			@Override
			public VisitorBag visitElement(T element){
				return visitor.visit(element);
			}
		};
	}

	public static JavaVisitor<JavaElement> allJavaElement(){
		return new JavaVisitor<JavaElement>(JavaElement.class){

			@Override
			public VisitorBag visitElement(JavaElement element){
				VisitorBag bag = new VisitorBag();
				bag.compute(element);
				return bag;
			}
			
		};
	}
	
	public static JavaClassBaseVisitor allClassBase(){
		return new JavaClassBaseVisitor();
	}
	
	public static JavaClassVisitor allClass(){
		return new JavaClassVisitor();
	}
	
	public static JavaInterfaceVisitor allInterface(){
		return new JavaInterfaceVisitor();
	}
	
	public static JavaEnumerationVisitor allEnum(){
		return new JavaEnumerationVisitor();
	}
	
	public static JavaAnnotationClassVisitor allAnnotationClass(){
		return new JavaAnnotationClassVisitor();
	}
	
	public static JavaAnnotationVisitor allAnnotation(){
		return new JavaAnnotationVisitor();
	}
	
	public static JavaVariableVisitor allVariable(){
		return new JavaVariableVisitor();
	}
	
	public static JavaFunctionVisitor allFunction(){
		return new JavaFunctionVisitor();
	}
	
	public static JavaValueVisitor allValue(){
		return new JavaValueVisitor();
	}
	
	public static JavaOperationVisitor allOperation(){
		return new JavaOperationVisitor();
	}
	
	public static JavaImportVisitor allImport(){
		return new JavaImportVisitor();
	}
	
	private Class<?> type;
	
	public JavaVisitor(Class<?> type){
		this.type = type;
	}
	
	@Override
	public boolean canVisit(Class<?> type){
//		System.out.println("can "+type+" visit "+this.type+" ? from "+Thread.currentThread().getStackTrace()[2]);
		return this.type.isAssignableFrom(type);
	}
	
	@Override
	public VisitorBag visit(JavaElement element){
		return visitElement((T) element);
	}
	
	public abstract VisitorBag visitElement(T element);
}